# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Partition` is a subset of the hardware resources of a :term:`CPC`
in DPM mode, virtualized as a separate computer.

Partitions can be created and deleted dynamically, and their resources such
as CPU, memory or I/O devices can be configured dynamically.
You can create as many partition definitions as you want, but only a specific
number of partitions can be active at any given time.

TODO: How can a user find out what the maximum is, before it is reached?

Partition resources are contained in CPC resources.

Partition resources only exist in CPCs that are in DPM mode. CPCs in classic
mode (or ensemble mode) have :term:`LPAR` resources, instead.
"""


import time
import copy
from requests.utils import quote

from ._manager import BaseManager
from ._resource import BaseResource
from ._exceptions import StatusTimeout
from ._nic import NicManager
from ._hba import HbaManager
from ._virtual_function import VirtualFunctionManager
from ._logging import logged_api_call
from ._utils import RC_PARTITION, make_query_str, datetime_from_timestamp, \
    timestamp_from_datetime

__all__ = ['PartitionManager', 'Partition']


class PartitionManager(BaseManager):
    """
    Manager providing access to the :term:`Partitions <Partition>` in a
    particular :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Cpc` object (in DPM mode):

    * :attr:`~zhmcclient.Cpc.partitions`

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, cpc):
        # This function should not go into the docs.
        # Parameters:
        #   cpc (:class:`~zhmcclient.Cpc`):
        #     CPC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',
            'status',
        ]

        super().__init__(
            resource_class=Partition,
            class_name=RC_PARTITION,
            session=cpc.manager.session,
            parent=cpc,
            base_uri='/api/partitions',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props,
            supports_properties=True)

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: :term:`CPC` defining the scope for this
        manager.
        """
        return self._parent

    @logged_api_call
    # pylint: disable=arguments-differ
    def list(self, full_properties=False, filter_args=None,
             additional_properties=None):
        """
        List the Partitions in this CPC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to any Partition to be included in the
          result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            This parameter requires HMC 2.16.0 or higher.

        Returns:

          : A list of :class:`~zhmcclient.Partition` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'partitions'
        list_uri = f'{self.cpc.uri}/partitions'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args,
            additional_properties)

    @logged_api_call(blanked_properties=['boot-ftp-password', 'ssc-master-pw'],
                     properties_pos=1)
    def create(self, properties):
        """
        Create and configure a Partition in this CPC.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this CPC.
        * Task permission to the "New Partition" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Partition' in the :term:`HMC API` book.

        Returns:

          Partition:
            The resource object for the new Partition.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.cpc.uri + '/partitions',
                                   body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        part = Partition(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return part


class Partition(BaseResource):
    """
    Representation of a :term:`Partition`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.PartitionManager`).

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.PartitionManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, PartitionManager), (
            f"Partition init: Expected manager type {PartitionManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)
        # The manager objects for child resources (with lazy initialization):
        self._nics = None
        self._hbas = None
        self._virtual_functions = None

    @property
    def nics(self):
        """
        :class:`~zhmcclient.NicManager`: Access to the :term:`NICs <NIC>` in
        this Partition.
        """
        # We do here some lazy loading.
        if not self._nics:
            self._nics = NicManager(self)
        return self._nics

    @property
    def hbas(self):
        """
        :class:`~zhmcclient.HbaManager`: Access to the :term:`HBAs <HBA>` in
        this Partition.

        On systems with the "dpm-storage-management"
        :ref:`firmware feature <firmware features>` (i.e. z14), the CPC will
        not have any HBA objects anymore (they are now Virtual Storage Resource
        objects), but this property still provides a manager object for
        consistency.
        """
        # We do here some lazy loading.
        if not self._hbas:
            self._hbas = HbaManager(self)
        return self._hbas

    @property
    def virtual_functions(self):
        """
        :class:`~zhmcclient.VirtualFunctionManager`: Access to the
        :term:`Virtual Functions <Virtual Function>` in this Partition.
        """
        # We do here some lazy loading.
        if not self._virtual_functions:
            self._virtual_functions = VirtualFunctionManager(self)
        return self._virtual_functions

    @logged_api_call
    def feature_enabled(self, feature_name):
        """
        Indicates whether the specified
        :ref:`firmware feature <firmware features>` is enabled for the CPC of
        this partition.

        The specified firmware feature must be available for the CPC.

        For a list of available firmware features, see section
        "Firmware Features" in the :term:`HMC API` book, or use the
        :meth:`feature_info` method.

        HMC/SE version requirements:

        * HMC version >= 2.14.0 with HMC API version >= 2.23

        Authorization requirements:

        * Object-access permission to this partition.

        Parameters:

          feature_name (:term:`string`): The name of the firmware feature.

        Returns:

          bool: `True` if the firmware feature is enabled, or `False` if the
          firmware feature is disabled.

        Raises:

          :exc:`ValueError`: Firmware features are not supported on the HMC.
          :exc:`ValueError`: The specified firmware feature is not available
            for the CPC.
          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        feature_list = self.prop('available-features-list', None)
        if feature_list is None:
            raise ValueError("Firmware features are not supported on the HMC")
        for feature in feature_list:
            if feature['name'] == feature_name:
                break
        else:
            raise ValueError(
                f"Firmware feature {feature_name} is not available for CPC "
                f"{self.manager.cpc.name}")
        return feature['state']  # pylint: disable=undefined-loop-variable

    @logged_api_call
    def feature_info(self):
        """
        Returns information about the :ref:`firmware features` available for
        the CPC of this partition.

        HMC/SE version requirements:

        * HMC version >= 2.14.0 with HMC API version >= 2.23

        Authorization requirements:

        * Object-access permission to this partition.

        Returns:

          :term:`iterable`:
            An iterable where each item represents one firmware feature that is
            available for the CPC of this partition.

            Each item is a dictionary with the following items:

            * `name` (:term:`unicode string`): Name of the feature.
            * `description` (:term:`unicode string`): Short description of
              the feature.
            * `state` (bool): Enablement state of the feature (`True` if
              enabled, `False` if disabled).

        Raises:

          :exc:`ValueError`: Firmware features are not supported on the HMC.
          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        feature_list = self.prop('available-features-list', None)
        if feature_list is None:
            raise ValueError("Firmware features are not supported on the HMC")
        return feature_list

    @logged_api_call
    def start(self, wait_for_completion=True, operation_timeout=None,
              status_timeout=None):
        """
        Start (activate) this Partition, using the HMC operation "Start
        Partition".

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it takes a few seconds until the partition
        status has reached the desired value (it still may show status
        "paused"). If `wait_for_completion=True`, this method repeatedly checks
        the status of the partition after the HMC operation has completed, and
        waits until the status is in one of the desired states "active" or
        "degraded".

        TODO: Describe what happens if the maximum number of active partitions
        is exceeded.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Start Partition" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the partition
            has reached the desired status, after the HMC operation has
            completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

        Returns:

          :class:`py:dict` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns an empty
            :class:`py:dict` object.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.

            This job supports cancellation. Note that it may no longer be
            possible to cancel the job after some point. The job status and
            reason codes will indicate whether the job was canceled or ran to
            completion.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired partition status.
        """
        result = self.manager.session.post(
            self.uri + '/operations/start', resource=self,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["active", "degraded"]
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def stop(self, wait_for_completion=True, operation_timeout=None,
             status_timeout=None):
        """
        Stop (deactivate) this Partition, using the HMC operation "Stop
        Partition".

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Stop Partition" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the partition
            has reached the desired status, after the HMC operation has
            completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

        Returns:

          :class:`py:dict` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns an empty
            :class:`py:dict` object.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.

            This job supports cancellation. Note that it may no longer be
            possible to cancel the job after some point. The job status and
            reason codes will indicate whether the job was canceled or ran to
            completion.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired partition status.
        """
        result = self.manager.session.post(
            self.uri + '/operations/stop', resource=self,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["stopped"]
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def delete(self):
        """
        Delete this Partition.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Delete Partition" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call(blanked_properties=['boot-ftp-password', 'ssc-master-pw'],
                     properties_pos=1)
    def update_properties(self, properties):
        """
        Update writeable properties of this Partition.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Partition Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'Partition object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    @logged_api_call
    def dump_partition(self, parameters, wait_for_completion=True,
                       operation_timeout=None):
        """
        Dump this Partition, by loading a standalone dump program from a SCSI
        device and starting its execution, using the HMC operation
        'Dump Partition'.

        HMC/SE version requirements:

        * SE version >= 2.13.1 without
          :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Dump Partition" task.

        Parameters:

          parameters (dict): Input parameters for the operation.
            Allowable input parameters are defined in section
            'Request body contents' in section 'Dump Partition' in the
            :term:`HMC API` book.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          :class:`py:dict` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns an empty
            :class:`py:dict` object.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        result = self.manager.session.post(
            self.uri + '/operations/scsi-dump', resource=self,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout,
            body=parameters)
        return result

    @logged_api_call
    def start_dump_program(self, parameters, wait_for_completion=True,
                           operation_timeout=None):
        """
        Dump this Partition, by loading a standalone dump program from a storage
        volume and starting its execution, using the HMC operation
        'Start Dump Program'.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Dump Partition" task.

        Parameters:

          parameters (dict): Input parameters for the operation.
            Allowable input parameters are defined in section
            'Request body contents' in section 'Start Dump Program' in the
            :term:`HMC API` book.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          :class:`py:dict` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns an empty
            :class:`py:dict` object.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        result = self.manager.session.post(
            self.uri + '/operations/start-dump-program', resource=self,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout,
            body=parameters)
        return result

    @logged_api_call
    def psw_restart(self, wait_for_completion=True, operation_timeout=None):
        """
        Initiates a PSW restart for this Partition, using the HMC operation
        'Perform PSW Restart'.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "PSW Restart" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          :class:`py:dict` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns an empty
            :class:`py:dict` object.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        result = self.manager.session.post(
            self.uri + '/operations/psw-restart', resource=self,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def mount_iso_image(self, image, image_name, ins_file_name):
        """
        Upload an ISO image and associate it to this Partition
        using the HMC operation 'Mount ISO Image'.

        When the partition already has an ISO image associated,
        the newly uploaded image replaces the current one.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Partition Details" task.

        Parameters:

          image (:term:`byte string` or file-like object):
            The content of the ISO image.

            Images larger than 2GB cannot be specified as a Byte string; they
            must be specified as a file-like object.

            File-like objects must have opened the file in binary mode.

          image_name (:term:`string`): The displayable name of the image.

            This value must be a valid Linux file name without directories,
            must not contain blanks, and must end with '.iso' in lower case.

            This value will be shown in the 'boot-iso-image-name' property of
            this partition.

          ins_file_name (:term:`string`): The path name of the INS file within
            the file system of the ISO image.

            This value will be shown in the 'boot-iso-ins-file' property of
            this partition.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms_str = (
            f"?image-name={quote(image_name, safe='')}"
            f"&ins-file-name={quote(ins_file_name, safe='')}")
        self.manager.session.post(
            self.uri + '/operations/mount-iso-image' + query_parms_str,
            resource=self, body=image)

    @logged_api_call
    def unmount_iso_image(self):
        """
        Unmount the currently mounted ISO from this Partition using the HMC
        operation 'Unmount ISO Image'. This operation sets the partition's
        'boot-iso-image-name' and 'boot-iso-ins-file' properties to null.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Partition Details" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        self.manager.session.post(
            self.uri + '/operations/unmount-iso-image', resource=self)

    @logged_api_call
    def open_os_message_channel(self, include_refresh_messages=True):
        """
        Open a JMS message channel to this partition's operating system,
        returning the string "topic" representing the message channel.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Operating System Messages" task at least
          in view-only mode.

        Parameters:

          include_refresh_messages (bool):
            Boolean controlling whether refresh operating systems messages
            should be sent, as follows:

            * If `True`, refresh messages will be recieved when the user
              connects to the topic. The default.

            * If `False`, refresh messages will not be recieved when the user
              connects to the topic.

        Returns:

          :term:`string`:

            Returns a string representing the os-message-notification JMS
            topic. The user can connect to this topic to start the flow of
            operating system messages.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'include-refresh-messages': include_refresh_messages}
        result = self.manager.session.post(
            self.uri + '/operations/open-os-message-channel', resource=self,
            body=body)
        return result['topic-name']

    @logged_api_call
    def send_os_command(self, os_command_text, is_priority=False):
        """
        Send a command to the operating system running in this partition.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Operating System Messages" task in
          modification mode.

        Parameters:

          os_command_text (string): The text of the operating system command.

          is_priority (bool):
            Boolean controlling whether this is a priority operating system
            command, as follows:

            * If `True`, this message is treated as a priority operating
              system command.

            * If `False`, this message is not treated as a priority
              operating system command. The default.

        Returns:

          None

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'is-priority': is_priority,
                'operating-system-command-text': os_command_text}
        self.manager.session.post(
            self.uri + '/operations/send-os-cmd', resource=self, body=body)

    @logged_api_call
    def list_os_messages(self, begin=None, end=None):
        """
        List all currently available operating system messages for this
        partition.

        Only a certain amount of OS message data from each partition is
        preserved by the HMC for retrieval by this operation. If the OS
        produces more than that amount, the oldest non-held, non-priority
        OS messages are no longer available. A gap in the sequence numbers
        indicates a loss of messages. A loss may be due to that space
        limitation, or it may be due to the deletion of messages by a console
        user or the OS.

        HMC/SE version requirements:

        * SE version >= 2.14.0

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Operating System Messages" task (optionally
          in view-only mode).

        Parameters:

          begin (integer): A message sequence number to limit returned
            messages. OS messages with a sequence number less than this are
            omitted from the results. If `None`, no such filtering is
            performed.

          end (integer): A message sequence number to limit returned
            messages. OS messages with a sequence number greater than this are
            omitted from the results. If `None`, no such filtering is
            performed.

        Returns:

          list of dict: List of OS messages, where each OS message is a dict
          with the items defined for the "os-message-info" data structure
          in the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms = []
        if begin is not None:
            query_parms.append(f'begin-sequence-number={begin}')
        if end is not None:
            query_parms.append(f'end-sequence-number={end}')
        query_str = make_query_str(query_parms)
        result = self.manager.session.get(
            f'{self.uri}/operations/list-os-messages{query_str}',
            resource=self)
        return result

    @logged_api_call
    def create_os_websocket(self, force_takeover=False):
        """
        Create a WebSocket on the HMC, which allows accessing the console of
        the operating system running in the partition using the integrated
        ASCII console of the HMC, and return the WebSocket URI for use by a
        WebSocket client.

        This is done by performing the "Get ASCII Console WebSocket URI"
        HMC operation.

        For more details on how to use a WebSocket client to interact with the
        integrated ASCII console, see
        :ref:`Using WebSocket to access OS console`.

        HMC/SE version requirements:

        * HMC version >= 2.14.0 with HMC API version >= 2.22

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Integrated ASCII Console" task.

        Parameters:

          force_takeover (bool):
            Boolean controlling whether to break any possibly existing
            WebSockets on other HMCs to the same partition, as follows:

            * If `True`, existing WebSockets are broken up and the operation
              proceeds.

            * If `False`, existing WebSockets are not broken up and the
              operation fails.

            Note that only existing WebSockets on *other* HMCs can be taken
            over, but not existing WebSockets on the current HMC.

        Returns:

          :term:`string`:

            Returns a string representing the canonical URI of the new
            WebSocket, e.g.
            ``/api/websock/4a4f1hj12hldmm26brcpfnydk663gt6gtyxq4iwto26g2r6wq1/1``.

            Depending on which WebSocket client is used, a full URI may need to
            be constructed from the returned string by prepending the secure
            WebSocket URI scheme ``wss`` and the HMC's IP address and port, e.g.
            ``wss://9.10.11.12:6794/api/websock/4a4f1hj12hldmm26brcpfnydk663gt6gtyxq4iwto26g2r6wq1/1``.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """  # pylint: disable=line-too-long
        body = {'force-takeover': force_takeover}
        result = self.manager.session.post(
            self.uri + '/operations/get-ascii-console-websocket-uri',
            resource=self, body=body)
        return result['websocket-uri']

    @logged_api_call
    def wait_for_status(self, status, status_timeout=None):
        """
        Wait until the status of this partition has a desired value.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Parameters:

          status (:term:`string` or iterable of :term:`string`):
            Desired partition status or set of status values to reach; one or
            more of the values defined for the 'status' property in the
            data model for partitions in the :term:`HMC API` book.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the partition
            has reached one of the desired status values. The special value 0
            means that no timeout is set.
            `None` means that the default status timeout will be used.
            If the timeout expires, a :exc:`~zhmcclient.StatusTimeout` is
            raised.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.StatusTimeout`: The status timeout expired while
            waiting for the desired partition status.
        """
        if status_timeout is None:
            status_timeout = \
                self.manager.session.retry_timeout_config.status_timeout
        if status_timeout > 0:
            end_time = time.time() + status_timeout
        if isinstance(status, (list, tuple)):
            statuses = status
        else:
            statuses = [status]

        while True:
            actual_status = self.get_properties_pulled('status')

            if actual_status in statuses:
                return

            # pylint: disable=possibly-used-before-assignment
            if status_timeout > 0 and time.time() > end_time:
                raise StatusTimeout(
                    f"Waiting for partition {self.name} to reach status(es) "
                    f"'{statuses}' timed out after {status_timeout} s - "
                    f"current status is '{actual_status}'",
                    actual_status, statuses, status_timeout)

            time.sleep(1)  # Avoid hot spin loop

    @logged_api_call
    def increase_crypto_config(self, crypto_adapters,
                               crypto_domain_configurations):
        """
        Add crypto adapters and/or crypto domains to the crypto configuration
        of this partition.

        The general principle for maintaining crypto configurations of
        partitions is as follows: Each adapter included in the crypto
        configuration of a partition has all crypto domains included in the
        crypto configuration. Each crypto domain included in the crypto
        configuration has the same access mode on all adapters included in the
        crypto configuration.

        Example: Assume that the current crypto configuration of a partition
        includes crypto adapter A and crypto domains 0 and 1. When this method
        is called to add adapter B and domain configurations for domains 1 and
        2, the resulting crypto configuration of the partition will include
        domains 0, 1, and 2 on each of the adapters A and B.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Object-access permission to the specified Crypto Adapter.
        * Task permission to the "Partition Details" task.

        Parameters:

          crypto_adapters (:term:`iterable` of :class:`~zhmcclient.Adapter`):
            Crypto adapters that should be added to the crypto configuration of
            this partition.

          crypto_domain_configurations (:term:`iterable` of `domain_config`):
            Crypto domain configurations that should be added to the crypto
            configuration of this partition.

            A crypto domain configuration (`domain_config`) is a dictionary
            with the following keys:

            * ``"domain-index"`` (:term:`integer`): Domain index of the crypto
              domain.

              The domain index is a number in the range of 0 to a maximum that
              depends on the model of the crypto adapter and the CPC model. For
              the Crypto Express 5S adapter in a z13, the maximum domain index
              is 84.

            * ``"access-mode"`` (:term:`string`): Access mode for the crypto
              domain.

              The access mode specifies the way the partition can use the
              crypto domain on the crypto adapter(s), using one of the
              following string values:

              * ``"control"`` - The partition can load cryptographic keys into
                the domain, but it may not use the domain to perform
                cryptographic operations.

              * ``"control-usage"`` - The partition can load cryptographic keys
                into the domain, and it can use the domain to perform
                cryptographic operations.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        crypto_adapter_uris = [a.uri for a in crypto_adapters]
        body = {'crypto-adapter-uris': crypto_adapter_uris,
                'crypto-domain-configurations': crypto_domain_configurations}
        self.manager.session.post(
            self.uri + '/operations/increase-crypto-configuration',
            resource=self, body=body)

    @logged_api_call
    def decrease_crypto_config(self, crypto_adapters,
                               crypto_domain_indexes):
        """
        Remove crypto adapters and/or crypto domains from the crypto
        configuration of this partition.

        For the general principle for maintaining crypto configurations of
        partitions, see :meth:`~zhmcclient.Partition.increase_crypto_config`.

        Example: Assume that the current crypto configuration of a partition
        includes crypto adapters A, B and C and crypto domains 0, 1, and 2 (on
        each of the adapters). When this method is called to remove adapter C
        and domain 2, the resulting crypto configuration of the partition will
        include domains 0 and 1 on each of the adapters A and B.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Object-access permission to the specified Crypto Adapters.
        * Task permission to the "Partition Details" task.

        Parameters:

          crypto_adapters (:term:`iterable` of :class:`~zhmcclient.Adapter`):
            Crypto adapters that should be removed from the crypto
            configuration of this partition.

          crypto_domain_indexes (:term:`iterable` of :term:`integer`):
            Domain indexes of the crypto domains that should be removed from
            the crypto configuration of this partition. For values, see
            :meth:`~zhmcclient.Partition.increase_crypto_config`.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        crypto_adapter_uris = [a.uri for a in crypto_adapters]
        body = {'crypto-adapter-uris': crypto_adapter_uris,
                'crypto-domain-indexes': crypto_domain_indexes}
        self.manager.session.post(
            self.uri + '/operations/decrease-crypto-configuration',
            resource=self, body=body)

    @logged_api_call
    def change_crypto_domain_config(self, crypto_domain_index, access_mode):
        """
        Change the access mode for a crypto domain that is currently included
        in the crypto configuration of this partition.

        The access mode will be changed for the specified crypto domain on all
        crypto adapters currently included in the crypto configuration of this
        partition.

        For the general principle for maintaining crypto configurations of
        partitions, see :meth:`~zhmcclient.Partition.increase_crypto_config`.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Partition Details" task.

        Parameters:

          crypto_domain_index (:term:`integer`):
            Domain index of the crypto domain to be changed. For values, see
            :meth:`~zhmcclient.Partition.increase_crypto_config`.

          access_mode (:term:`string`):
            The new access mode for the crypto domain. For values, see
            :meth:`~zhmcclient.Partition.increase_crypto_config`.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'domain-index': crypto_domain_index,
                'access-mode': access_mode}
        self.manager.session.post(
            self.uri + '/operations/change-crypto-domain-configuration',
            resource=self, body=body)

    @logged_api_call
    def zeroize_crypto_domain(self, crypto_adapter, crypto_domain_index):
        """
        Zeroize a single crypto domain on a crypto adapter.

        Zeroizing a crypto domain clears the cryptographic keys and
        non-compliance mode settings in the crypto domain.

        The crypto domain must be attached to this partition in "control-usage"
        access mode.

        Supported CPC versions: z14 GA2 and above, and the corresponding
        LinuxOne systems.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Object-access permission to the specified Crypto Adapter.
        * Task permission to the "Zeroize Crypto Domain" task.

        Parameters:

          crypto_adapter (:class:`~zhmcclient.Adapter`):
            Crypto adapter with the crypto domain to be zeroized.

          crypto_domain_index (:term:`integer`):
            Domain index of the crypto domain to be zeroized.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {
            'crypto-adapter-uri': crypto_adapter.uri,
            'domain-index': crypto_domain_index
        }
        self.manager.session.post(
            self.uri + '/operations/zeroize-crypto-domain', resource=self,
            body=body)

    @logged_api_call
    def attach_storage_group(self, storage_group):
        """
        Attach a :term:`storage group` to this partition.

        This will cause the :term:`storage volumes <storage volume>` of the
        storage group to be attached to the partition, instantiating any
        necessary :term:`virtual storage resource` objects.

        A storage group can be attached to a partition regardless of its
        fulfillment state. The fulfillment state of its storage volumes
        and thus of the entire storage group changes as volumes are discovered
        by DPM, and will eventually reach "complete".

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this partition.
        * Object-access permission to the specified storage group.
        * Task permission to the "Partition Details" task.

        Parameters:

          storage_group (:class:`~zhmcclient.StorageGroup`):
            Storage group to be attached. The storage group must not currently
            be attached to this partition.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'storage-group-uri': storage_group.uri}
        self.manager.session.post(
            self.uri + '/operations/attach-storage-group', resource=self,
            body=body)

    @logged_api_call
    def detach_storage_group(self, storage_group):
        """
        Detach a :term:`storage group` from this partition.

        This will cause the :term:`storage volumes <storage volume>` of the
        storage group to be detached from the partition, removing any
        :term:`virtual storage resource` objects that had been created upon
        attachment.

        A storage group can be detached from a partition regardless of its
        fulfillment state. The fulfillment state of its storage volumes
        changes as volumes are discovered by DPM.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this partition.
        * Task permission to the "Partition Details" task.

        Parameters:

          storage_group (:class:`~zhmcclient.StorageGroup`):
            Storage group to be detached. The storage group must currently
            be attached to this partition.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'storage-group-uri': storage_group.uri}
        self.manager.session.post(
            self.uri + '/operations/detach-storage-group', resource=self,
            body=body)

    @logged_api_call
    def list_attached_storage_groups(self, full_properties=False):
        """
        Return the storage groups that are attached to this partition.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this partition.

        Parameters:

          full_properties (bool):
            Controls that the full set of resource properties for each returned
            storage group is being retrieved, vs. only the following short set:
            "object-uri", "object-id", "class", "parent".

            TODO: Verify short list of properties.

        Returns:

          List of :class:`~zhmcclient.StorageGroup` objects representing the
          storage groups that are attached to this partition.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        sg_list = []
        sg_uris = self.get_property('storage-group-uris')
        if sg_uris:
            console = self.manager.cpc.manager.console
            for sg_uri in sg_uris:
                sg = console.storage_groups.resource_object(sg_uri)
                sg_list.append(sg)
                if full_properties:
                    sg.pull_full_properties()
        return sg_list

    @logged_api_call
    def assign_certificate(self, certificate):
        """
        Assigns a :term:`Certificate` to this partition.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to this partition.
        * Object-access permission to the specified certificate.
        * Task permission to the "Assign Secure Boot Certificates" task.

        Parameters:

          certificate (:class:`~zhmcclient.Certificate`):
            Certificate to be assigned. The certificate must not currently
            be assigned to this partition.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'certificate-uri': certificate.uri}
        self.manager.session.post(
            self.uri + '/operations/assign-certificate', resource=self,
            body=body)

    @logged_api_call
    def unassign_certificate(self, certificate):
        """
        Unassign a :term:`Certificate` from this partition.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to this partition.
        * Object-access permission to the specified certificate.
        * Task permission to the "Assign Secure Boot Certificates" task.

        Parameters:

          certificate (:class:`~zhmcclient.Certificate`):
            Certificate to be unassigned. The certificate must currently be
            assigned to this partition.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'certificate-uri': certificate.uri}
        self.manager.session.post(
            self.uri + '/operations/unassign-certificate', resource=self,
            body=body)

    def dump(self):
        """
        Dump this Partition resource with its properties and child resources
        (recursively) as a resource definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},

                # Child resources:
                "nics": [...],
                "hbas": [...],
                "virtual_functions": [...],
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        # Dump the child resources
        nics = self.nics.dump()
        if nics:
            resource_dict['nics'] = nics
        if self.hbas is not None:
            # Note: z14 and later do not have HBAs
            hbas = self.hbas.dump()
            if hbas:
                resource_dict['hbas'] = hbas
        virtual_functions = self.virtual_functions.dump()
        if virtual_functions:
            resource_dict['virtual_functions'] = virtual_functions

        return resource_dict

    @logged_api_call
    def get_sustainability_data(
            self, range="last-week", resolution="one-hour",
            custom_range_start=None, custom_range_end=None):
        # pylint: disable=redefined-builtin
        """
        Get energy management related metrics for the partition on a specific
        historical time range. The metrics are returned as multiple data points
        covering the requested time range with the requested resolution.
        This method performs the "Get Partition Historical Sustainability Data"
        HMC operation.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "environmental-metrics"

        Authorization requirements:

        * Object-access permission to this partition
        * Task permission to the "Environmental Dashboard" task

        Parameters:

          range (:term:`string`):
            Time range for the requested data points, as follows:

            * "last-day" - Last 24 hours.
            * "last-week" - Last 7 days (default).
            * "last-month" - Last 30 days.
            * "last-three-months" - Last 90 days.
            * "last-six-months" - Last 180 days.
            * "last-year" - Last 365 days.
            * "custom" - From `custom_range_start` to `custom_range_end`.

          resolution (:term:`string`):
            Resolution for the requested data points. This is the time interval
            in between the data points. For systems where the
            "environmental-metrics" API feature is not available, the minimum
            resolution is "one-hour".

            The possible values are as follows:

            * "fifteen-minutes" - 15 minutes.
            * "one-hour" - 60 minutes (default).
            * "one-day" - 24 hours.
            * "one-week" - 7 days.
            * "one-month" - 30 days.

          custom_range_start (:class:`~py:datetime.datetime`):
            Start of custom time range. Timezone-naive values are interpreted
            using the local system time. Required if `range` is "custom".

          custom_range_end (:class:`~py:datetime.datetime`):
            End of custom time range. Timezone-naive values are interpreted
            using the local system time. Required if `range` is "custom".

        Returns:

          dict: A dictionary with items as described for the response body
          of the "Get Partition Historical Sustainability Data" HMC operation.
          Timestamp fields are represented as timezone-aware
          :class:`~py:datetime.datetime` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {
            'range': range,
            'resolution': resolution,
        }
        if range == "custom":
            body['custom-range-start'] = \
                timestamp_from_datetime(custom_range_start)
            body['custom-range-end'] = \
                timestamp_from_datetime(custom_range_end)
        result = self.manager.session.post(
            self.uri + '/operations/get-historical-sustainability-data',
            body=body)
        for field_array in result.values():
            for item in field_array:
                if 'timestamp' in item:
                    item['timestamp'] = \
                        datetime_from_timestamp(item['timestamp'])
        return result



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _cpc module.
"""


import re
import copy
import pytest
import requests_mock

from zhmcclient import Client, Cpc, HTTPError, STPNode, Job
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources

# pylint: disable=unused-import
from tests.common.http_mocked_fixtures import http_mocked_session  # noqa: F401
from tests.common.http_mocked_fixtures import http_mocked_cpc_dpm  # noqa: F401
# pylint: enable=unused-import

# Object IDs and names of our faked CPCs:
CPC1_NAME = 'cpc 1'  # z13s in DPM mode
CPC1_OID = 'cpc1-oid'
CPC1_MAX_CRYPTO_DOMAINS = 40  # Crypto Express5S on a z13s
CPC2_NAME = 'cpc 2'  # z13s in classic mode
CPC2_OID = 'cpc2-oid'
CPC3_NAME = 'cpc 3'  # zEC12
CPC3_OID = 'cpc3-oid'
CPC4_NAME = 'cpc 4'  # z14-ZR1 in DPM mode
CPC4_OID = 'cpc4-oid'

# Additional CPC properties for temporary capacity tests
CPC1_CAPACITY_PROPS = {
    'is-on-off-cod-installed': True,
    'is-on-off-cod-enabled': True,
    'is-on-off-cod-activated': False,
    'on-off-cod-activation-date': None,
    'software-model-permanent': '700',
    'software-model-permanent-plus-billable': '700',
    'software-model-permanent-plus-temporary': None,
    # 'msu-permanent': TBD,
    # 'msu-permanen-plus-billable': TBD,
    # 'msu-permanent-plus-temporary': TBD,
    'processor-count-general-purpose': 7,
    'processor-count-service-assist': 6,
    'processor-count-aap': 5,
    'processor-count-ifl': 4,
    'processor-count-icf': 3,
    'processor-count-iip': 2,
    'processor-count-cbp2': 1,
}

HTTPError_400_7 = HTTPError({'http-status': 400, 'reason': 7})
HTTPError_404_1 = HTTPError({'http-status': 404, 'reason': 1})
HTTPError_409_1 = HTTPError({'http-status': 409, 'reason': 1})
HTTPError_409_4 = HTTPError({'http-status': 409, 'reason': 4})
HTTPError_409_5 = HTTPError({'http-status': 409, 'reason': 5})

# Names of our faked crypto adapters:
CRYPTO1_NAME = 'crypto 1'
CRYPTO2_NAME = 'crypto 2'

CPC1_UNUSED_CRYPTO_DOMAINS = list(range(4, CPC1_MAX_CRYPTO_DOMAINS))

GET_FREE_CRYPTO_DOMAINS_ENVIRONMENTS = {
    'env0-example': {
        'desc': "The example from the description of method "
                "Cpc.get_free_crypto_domains()",
        'cpc_name': CPC1_NAME,
        'adapter_names': [
            CRYPTO1_NAME,
            CRYPTO2_NAME,
        ],
        'partitions': [
            {
                'name': 'part-A',
                'adapter_names': [
                    CRYPTO1_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 0, 'access-mode': 'control-usage'},
                    {'domain-index': 1, 'access-mode': 'control'},
                    # We leave domain index 4 and higher untouched
                ],
            },
            {
                'name': 'part-B',
                'adapter_names': [
                    CRYPTO2_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 0, 'access-mode': 'control'},
                    {'domain-index': 1, 'access-mode': 'control-usage'},
                    {'domain-index': 2, 'access-mode': 'control-usage'},
                    # We leave domain index 4 and higher untouched
                ],
            },
            {
                'name': 'part-C',
                'adapter_names': [
                    CRYPTO1_NAME,
                    CRYPTO2_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 0, 'access-mode': 'control'},
                    {'domain-index': 1, 'access-mode': 'control'},
                    {'domain-index': 3, 'access-mode': 'control-usage'},
                    # We leave domain index 4 and higher untouched
                ],
            },
        ]
    },
    'env1-ocdu': {
        'desc': "Overlapped control of domains, but disjoint usage "
                "on all adapters",
        'cpc_name': CPC1_NAME,
        'adapter_names': [
            CRYPTO1_NAME,
            CRYPTO2_NAME,
        ],
        'partitions': [
            {
                'name': 'part-0',
                'adapter_names': [
                    CRYPTO1_NAME,
                    CRYPTO2_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 0, 'access-mode': 'control-usage'},
                    {'domain-index': 1, 'access-mode': 'control'},
                    {'domain-index': 2, 'access-mode': 'control'},
                    {'domain-index': 3, 'access-mode': 'control'},
                    # We leave domain index 4 and higher untouched
                ],
            },
            {
                'name': 'part-1',
                'adapter_names': [
                    CRYPTO1_NAME,
                    CRYPTO2_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 1, 'access-mode': 'control-usage'},
                    # We leave domain index 4 and higher untouched
                ],
            },
        ]
    },
    'env2-dcdu': {
        'desc': "Disjoint control and usage of domains on all adapters",
        'cpc_name': CPC1_NAME,
        'adapter_names': [
            CRYPTO1_NAME,
            CRYPTO2_NAME,
        ],
        'partitions': [
            {
                'name': 'part-0',
                'adapter_names': [
                    CRYPTO1_NAME,
                    CRYPTO2_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 0, 'access-mode': 'control-usage'},
                    {'domain-index': 1, 'access-mode': 'control'},
                    # We leave domain index 4 and higher untouched
                ],
            },
            {
                'name': 'part-2',
                'adapter_names': [
                    CRYPTO1_NAME,
                    CRYPTO2_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 2, 'access-mode': 'control-usage'},
                    {'domain-index': 3, 'access-mode': 'control'},
                    # We leave domain index 4 and higher untouched
                ],
            },
        ]
    },
    'env3-dcou': {
        'desc': "Disjoint control of domains, but overlapping usage on all "
                "adapters (this prevents activating the partitions at the "
                "same time)",
        'cpc_name': CPC1_NAME,
        'adapter_names': [
            CRYPTO1_NAME,
            CRYPTO2_NAME,
        ],
        'partitions': [
            {
                'name': 'part-1',
                'adapter_names': [
                    CRYPTO1_NAME,
                    CRYPTO2_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 1, 'access-mode': 'control-usage'},
                    {'domain-index': 2, 'access-mode': 'control'},
                    # We leave domain index 4 and higher untouched
                ],
            },
            {
                'name': 'part-2',
                'adapter_names': [
                    CRYPTO1_NAME,
                    CRYPTO2_NAME,
                ],
                'domain_configs': [
                    {'domain-index': 1, 'access-mode': 'control-usage'},
                    {'domain-index': 3, 'access-mode': 'control'},
                    # We leave domain index 4 and higher untouched
                ],
            },
        ]
    },
}

GET_FREE_CRYPTO_DOMAINS_SUCCESS_TESTCASES = [
    # (env_name, parm_adapter_names, exp_free_domains)

    # test cases for environment 'env0-example':
    (
        'env0-example',
        [],
        None,
    ),
    (
        'env0-example',
        [CRYPTO1_NAME],
        [1, 2] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env0-example',
        [CRYPTO2_NAME],
        [0] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env0-example',
        [CRYPTO1_NAME, CRYPTO2_NAME],
        [] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env0-example',
        None,
        [] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),

    # test cases for environment 'env1-ocdu':
    (
        'env1-ocdu',
        [],
        None,
    ),
    (
        'env1-ocdu',
        [CRYPTO1_NAME],
        [2, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env1-ocdu',
        [CRYPTO2_NAME],
        [2, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env1-ocdu',
        [CRYPTO1_NAME, CRYPTO2_NAME],
        [2, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env1-ocdu',
        None,
        [2, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),

    # test cases for environment 'env2-dcdu':
    (
        'env2-dcdu',
        [],
        None,
    ),
    (
        'env2-dcdu',
        [CRYPTO1_NAME],
        [1, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env2-dcdu',
        [CRYPTO2_NAME],
        [1, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env2-dcdu',
        [CRYPTO1_NAME, CRYPTO2_NAME],
        [1, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env2-dcdu',
        None,
        [1, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),

    # test cases for environment 'env3-dcou':
    (
        'env3-dcou',
        [],
        None,
    ),
    (
        'env3-dcou',
        [CRYPTO1_NAME],
        [0, 2, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env3-dcou',
        [CRYPTO2_NAME],
        [0, 2, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env3-dcou',
        [CRYPTO1_NAME, CRYPTO2_NAME],
        [0, 2, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
    (
        'env3-dcou',
        None,
        [0, 2, 3] + CPC1_UNUSED_CRYPTO_DOMAINS,
    ),
]

SET_AUTO_START_LIST_ENVIRONMENTS = {
    'env1': {
        'desc': "CPC with two partitions",
        'cpc_name': CPC1_NAME,
        'partitions': [
            {
                'name': 'part-A',
            },
            {
                'name': 'part-B',
            },
        ]
    },
}


def updated(dict1, dict2):
    """
    Return a copy of dict1 that is updated with dict2. This allows specifying
    testcase structures more easily.
    """
    dictres = dict1.copy()
    dictres.update(dict2)
    return dictres


class TestCpc:
    """All tests for the Cpc and CpcManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

    def add_cpc(self, cpc_name):
        """Add a faked CPC."""

        if cpc_name == CPC1_NAME:
            faked_cpc = self.session.hmc.cpcs.add({
                'object-id': CPC1_OID,
                # object-uri is set up automatically
                'parent': None,
                'class': 'cpc',
                'name': CPC1_NAME,
                'description': 'CPC #1 (z13s in DPM mode)',
                'status': 'active',
                'dpm-enabled': True,
                'is-ensemble-member': False,
                'iml-mode': 'dpm',
                'machine-type': '2965',
                'machine-model': 'N10',
            })
        elif cpc_name == CPC2_NAME:
            faked_cpc = self.session.hmc.cpcs.add({
                'object-id': CPC2_OID,
                # object-uri is set up automatically
                'parent': None,
                'class': 'cpc',
                'name': CPC2_NAME,
                'description': 'CPC #2 (z13s in classic mode)',
                'status': 'operating',
                'dpm-enabled': False,
                'is-ensemble-member': False,
                'iml-mode': 'lpar',
                'machine-type': '2965',
                'machine-model': 'N10',
            })
        elif cpc_name == CPC3_NAME:
            faked_cpc = self.session.hmc.cpcs.add({
                'object-id': CPC3_OID,
                # object-uri is set up automatically
                'parent': None,
                'class': 'cpc',
                'name': CPC3_NAME,
                'description': 'CPC #3 (zEC12)',
                'status': 'operating',
                # zEC12 does not have a dpm-enabled property
                'is-ensemble-member': False,
                'iml-mode': 'lpar',
                'machine-type': '2827',
                'machine-model': 'H43',
            })
        elif cpc_name == CPC4_NAME:
            faked_cpc = self.session.hmc.cpcs.add({
                'object-id': CPC4_OID,
                # object-uri is set up automatically
                'parent': None,
                'class': 'cpc',
                'name': CPC4_NAME,
                'description': 'CPC #4 (z14-ZR1 in DPM mode)',
                'status': 'active',
                'dpm-enabled': True,
                'is-ensemble-member': False,
                'iml-mode': 'dpm',
                'machine-type': '3607',
                'machine-model': 'LR1',
                'available-features-list': [],
            })
        else:
            raise ValueError(f"Invalid value for cpc_name: {cpc_name}")
        return faked_cpc

    @staticmethod
    def add_crypto_adapter(faked_cpc, adapter_name):
        """Add a faked crypto adapter to a faked CPC."""

        if adapter_name == CRYPTO1_NAME:
            faked_crypto_adapter = faked_cpc.adapters.add({
                'object-id': adapter_name + '-oid',
                # object-uri is automatically set
                'parent': faked_cpc.uri,
                'class': 'adapter',
                'name': adapter_name,
                'status': 'active',
                'type': 'crypto',
                'adapter-family': 'crypto',
                'detected-card-type': 'crypto-express-5s',
                'crypto-type': 'ep11-coprocessor',
                'crypto-number': 1,
                'adapter-id': '12C',
            })
        elif adapter_name == CRYPTO2_NAME:
            faked_crypto_adapter = faked_cpc.adapters.add({
                'object-id': adapter_name + '-oid',
                # object-uri is automatically set
                'parent': faked_cpc.uri,
                'class': 'adapter',
                'name': adapter_name,
                'status': 'active',
                'type': 'crypto',
                'adapter-family': 'crypto',
                'detected-card-type': 'crypto-express-5s',
                'crypto-type': 'cca-coprocessor',
                'crypto-number': 2,
                'adapter-id': '12D',
            })
        else:
            raise ValueError(f"Invalid value for crypto_name: {adapter_name}")
        return faked_crypto_adapter

    @staticmethod
    def add_partition(faked_cpc, part_name):
        """Add a faked partition to a faked CPC, with standard properties."""

        faked_partition = faked_cpc.partitions.add({
            'object-id': part_name + '-oid',
            # object-uri is automatically set
            'parent': faked_cpc.uri,
            'class': 'partition',
            'name': part_name,
            'status': 'active',
            'ifl-processors': 2,
            'initial-memory': 1024,
            'maximum-memory': 1024,
        })
        return faked_partition

    def test_cpcmanager_initial_attrs(self):
        """Test initial attributes of CpcManager."""

        cpc_mgr = self.client.cpcs

        # Verify all public properties of the manager object
        assert cpc_mgr.resource_class is Cpc
        assert cpc_mgr.session is self.session
        assert cpc_mgr.parent is None
        assert cpc_mgr.client is self.client

    # TODO: Test for CpcManager.__repr__()

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            ({},
             ['object-uri', 'name', 'status']),
            (dict(full_properties=False),
             ['object-uri', 'name', 'status']),
            (dict(full_properties=True),
             None),
        ]
    )
    def test_cpcmanager_list_full_properties(
            self, full_properties_kwargs, prop_names):
        """Test CpcManager.list() with full_properties."""

        # Add two faked CPCs
        faked_cpc1 = self.add_cpc(CPC1_NAME)
        faked_cpc2 = self.add_cpc(CPC2_NAME)

        exp_faked_cpcs = [faked_cpc1, faked_cpc2]
        cpc_mgr = self.client.cpcs

        # Execute the code to be tested
        cpcs = cpc_mgr.list(**full_properties_kwargs)

        assert_resources(cpcs, exp_faked_cpcs, prop_names)

    @pytest.mark.parametrize(
        "filter_args, exp_names", [
            ({'object-id': CPC1_OID},
             [CPC1_NAME]),
            ({'object-id': CPC2_OID},
             [CPC2_NAME]),
            ({'object-id': [CPC1_OID, CPC2_OID]},
             [CPC1_NAME, CPC2_NAME]),
            ({'object-id': [CPC1_OID, CPC1_OID]},
             [CPC1_NAME]),
            ({'object-id': CPC1_OID + 'foo'},
             []),
            ({'object-id': [CPC1_OID, CPC2_OID + 'foo']},
             [CPC1_NAME]),
            ({'object-id': [CPC2_OID + 'foo', CPC1_OID]},
             [CPC1_NAME]),
            ({'name': CPC1_NAME},
             [CPC1_NAME]),
            ({'name': CPC2_NAME},
             [CPC2_NAME]),
            ({'name': [CPC1_NAME, CPC2_NAME]},
             [CPC1_NAME, CPC2_NAME]),
            ({'name': CPC1_NAME + 'foo'},
             []),
            ({'name': [CPC1_NAME, CPC2_NAME + 'foo']},
             [CPC1_NAME]),
            ({'name': [CPC2_NAME + 'foo', CPC1_NAME]},
             [CPC1_NAME]),
            ({'name': [CPC1_NAME, CPC1_NAME]},
             [CPC1_NAME]),
            ({'name': '.*cpc 1'},
             [CPC1_NAME]),
            ({'name': 'cpc 1.*'},
             [CPC1_NAME]),
            ({'name': 'cpc .'},
             [CPC1_NAME, CPC2_NAME]),
            ({'name': '.pc 1'},
             [CPC1_NAME]),
            ({'name': '.+'},
             [CPC1_NAME, CPC2_NAME]),
            ({'name': 'cpc 1.+'},
             []),
            ({'name': '.+cpc 1'},
             []),
            ({'name': CPC1_NAME,
              'object-id': CPC1_OID},
             [CPC1_NAME]),
            ({'name': CPC1_NAME,
              'object-id': CPC1_OID + 'foo'},
             []),
            ({'name': CPC1_NAME + 'foo',
              'object-id': CPC1_OID},
             []),
            ({'name': CPC1_NAME + 'foo',
              'object-id': CPC1_OID + 'foo'},
             []),
        ]
    )
    def test_cpcmanager_list_filter_args(self, filter_args, exp_names):
        """Test CpcManager.list() with filter_args."""

        # Add two faked CPCs
        self.add_cpc(CPC1_NAME)
        self.add_cpc(CPC2_NAME)

        cpc_mgr = self.client.cpcs

        # Execute the code to be tested
        cpcs = cpc_mgr.list(filter_args=filter_args)

        assert len(cpcs) == len(exp_names)
        if exp_names:
            names = [ad.properties['name'] for ad in cpcs]
            assert set(names) == set(exp_names)

    # TODO: Test for initial Cpc attributes (lpars, partitions, adapters,
    #       virtual_switches, reset_activation_profiles,
    #       image_activation_profiles, load_activation_profiles, )

    def test_cpc_repr(self):
        """Test Cpc.__repr__()."""

        # Add a faked CPC
        faked_cpc = self.add_cpc(CPC1_NAME)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=faked_cpc.name)

        # Execute the code to be tested
        repr_str = repr(cpc)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{cpc.__class__.__name__}\s+at\s+'
            rf'0x{id(cpc):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "cpc_name, exp_dpm_enabled", [
            (CPC1_NAME, True),
            (CPC2_NAME, False),
            (CPC3_NAME, False),
        ]
    )
    def test_cpc_dpm_enabled(self, cpc_name, exp_dpm_enabled):
        """Test Cpc.dpm_enabled."""

        # Add a faked CPC
        self.add_cpc(cpc_name)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        # Execute the code to be tested
        dpm_enabled = cpc.dpm_enabled

        assert dpm_enabled == exp_dpm_enabled

    # TODO: Test for Cpc.maximum_active_partitions

    @pytest.mark.parametrize(
        "desc, cpc_name, available_features, feature_name, "
        "exp_feature_enabled, exp_exc", [
            (
                "No feature support on the CPC",
                CPC1_NAME,
                None,
                'fake-feature1', None, ValueError()
            ),
            (
                "Feature not available on the CPC (empty feature list)",
                CPC4_NAME,
                [],
                'fake-feature1', None, ValueError()
            ),
            (
                "Feature not available on the CPC (one other feature avail)",
                CPC4_NAME,
                [
                    dict(name='fake-feature-foo', state=True),
                ],
                'fake-feature1', None, ValueError()
            ),
            (
                "Feature disabled (the only feature available)",
                CPC4_NAME,
                [
                    dict(name='fake-feature1', state=False),
                ],
                'fake-feature1', False, None
            ),
            (
                "Feature enabled (the only feature available)",
                CPC4_NAME,
                [
                    dict(name='fake-feature1', state=True),
                ],
                'fake-feature1', True, None
            ),
        ]
    )
    def test_cpc_feature_enabled(
            self, desc, cpc_name, available_features, feature_name,
            exp_feature_enabled, exp_exc):
        # pylint: disable=unused-argument
        """Test Cpc.feature_enabled()."""

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)

        # Set up the firmware feature list
        if available_features is not None:
            faked_cpc.properties['available-features-list'] = \
                available_features

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        if exp_exc:
            with pytest.raises(exp_exc.__class__):

                # Execute the code to be tested
                cpc.feature_enabled(feature_name)

        else:

            # Execute the code to be tested
            act_feature_enabled = cpc.feature_enabled(feature_name)

            assert act_feature_enabled == exp_feature_enabled

    @pytest.mark.parametrize(
        "desc, cpc_name, available_features, exp_exc", [
            (
                "No feature support on the CPC",
                CPC1_NAME,
                None,
                ValueError()
            ),
            (
                "Feature not available on the CPC (empty feature list)",
                CPC4_NAME,
                [],
                None
            ),
            (
                "Feature not available on the CPC (one other feature avail)",
                CPC4_NAME,
                [
                    dict(name='fake-feature-foo', state=True),
                ],
                None
            ),
            (
                "Feature disabled (the only feature available)",
                CPC4_NAME,
                [
                    dict(name='fake-feature1', state=False),
                ],
                None
            ),
            (
                "Feature enabled (the only feature available)",
                CPC4_NAME,
                [
                    dict(name='fake-feature1', state=True),
                ],
                None
            ),
        ]
    )
    def test_cpc_feature_info(
            self, desc, cpc_name, available_features, exp_exc):
        # pylint: disable=unused-argument
        """Test Cpc.feature_info()."""

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)

        # Set up the firmware feature list
        if available_features is not None:
            faked_cpc.properties['available-features-list'] = \
                available_features

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        if exp_exc:
            with pytest.raises(exp_exc.__class__):

                # Execute the code to be tested
                cpc.feature_info()

        else:

            # Execute the code to be tested
            act_features = cpc.feature_info()

            assert act_features == available_features

    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New CPC description'},
            {'acceptable-status': ['active', 'degraded'],
             'description': 'New CPC description'},
        ]
    )
    def test_cpc_update_properties(self, input_props):
        """Test Cpc.update_properties()."""

        # Add a faked CPC
        faked_cpc = self.add_cpc(CPC1_NAME)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=faked_cpc.name)

        cpc.pull_full_properties()
        saved_properties = copy.deepcopy(cpc.properties)

        # Execute the code to be tested
        cpc.update_properties(properties=input_props)

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in cpc.properties
            prop_value = cpc.properties[prop_name]
            assert prop_value == exp_prop_value

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        cpc.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in cpc.properties
            prop_value = cpc.properties[prop_name]
            assert prop_value == exp_prop_value

    @pytest.mark.parametrize(
        "wait_for_completion", [True]
    )
    @pytest.mark.parametrize(
        "cpc_name, initial_status, exp_status, exp_error", [
            (CPC1_NAME, 'not-operating', 'active', None),
            (CPC2_NAME, 'not-operating', None, HTTPError_409_5),
            (CPC3_NAME, 'not-operating', None, HTTPError_409_5),
        ]
    )
    def test_cpc_start(self, cpc_name, initial_status, exp_status, exp_error,
                       wait_for_completion):
        """Test Cpc.start()."""

        # wait_for_completion=False not implemented in mock support:
        assert wait_for_completion is True

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)

        # Set initial status of the CPC for this test
        faked_cpc.properties['status'] = initial_status

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                result = cpc.start(wait_for_completion=wait_for_completion)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            result = cpc.start(wait_for_completion=wait_for_completion)

            if wait_for_completion:
                assert result is None
            else:
                raise NotImplementedError

            cpc.pull_full_properties()
            status = cpc.properties['status']
            assert status == exp_status

    @pytest.mark.parametrize(
        "wait_for_completion", [True]
    )
    @pytest.mark.parametrize(
        "cpc_name, initial_status, exp_status, exp_error", [
            (CPC1_NAME, 'active', 'not-operating', None),
            (CPC2_NAME, 'operating', None, HTTPError_409_5),
            (CPC3_NAME, 'operating', None, HTTPError_409_5),
        ]
    )
    def test_cpc_stop(self, cpc_name, initial_status, exp_status, exp_error,
                      wait_for_completion):
        """Test Cpc.stop()."""

        # wait_for_completion=False not implemented in mock support:
        assert wait_for_completion is True

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)

        # Set initial status of the CPC for this test
        faked_cpc.properties['status'] = initial_status

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                result = cpc.stop(wait_for_completion=wait_for_completion)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            result = cpc.stop(wait_for_completion=wait_for_completion)

            if wait_for_completion:
                assert result is None
            else:
                raise NotImplementedError

            cpc.pull_full_properties()
            status = cpc.properties['status']
            assert status == exp_status

    @pytest.mark.parametrize(
        "wait_for_completion", [True]
    )
    @pytest.mark.parametrize(
        "cpc_name, initial_status, force, exp_status, exp_error", [
            # Different operational modes
            (CPC1_NAME, 'not-operating', False, None, HTTPError_409_4),
            (CPC2_NAME, 'not-operating', False, 'operating', None),
            # Different initial inactive statuses
            (CPC3_NAME, 'not-operating', False, 'operating', None),
            (CPC3_NAME, 'not-operating', True, 'operating', None),
            (CPC3_NAME, 'no-power', False, 'operating', None),
            (CPC3_NAME, 'no-power', True, 'operating', None),
            # Different initial active statuses
            (CPC3_NAME, 'operating', False, None, HTTPError_409_1),
            (CPC3_NAME, 'operating', True, 'operating', None),
            (CPC3_NAME, 'degraded', False, None, HTTPError_409_1),
            (CPC3_NAME, 'degraded', True, 'operating', None),
            (CPC3_NAME, 'acceptable', False, None, HTTPError_409_1),
            (CPC3_NAME, 'acceptable', True, 'operating', None),
            (CPC3_NAME, 'exceptions', False, None, HTTPError_409_1),
            (CPC3_NAME, 'exceptions', True, 'operating', None),
            (CPC3_NAME, 'service-required', False, None, HTTPError_409_1),
            (CPC3_NAME, 'service-required', True, 'operating', None),
            # Different initial bad statuses
            (CPC3_NAME, 'not-communicating', False, None, HTTPError_409_1),
            (CPC3_NAME, 'not-communicating', True, None, HTTPError_409_1),
            (CPC3_NAME, 'status-check', False, None, HTTPError_409_1),
            (CPC3_NAME, 'status-check', True, None, HTTPError_409_1),
        ]
    )
    def test_cpc_activate(
            self, cpc_name, initial_status, force, exp_status, exp_error,
            wait_for_completion):
        """Test Cpc.activate()."""

        # wait_for_completion=False not implemented in mock support:
        assert wait_for_completion is True

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)

        # Set initial status of the CPC for this test
        faked_cpc.properties['status'] = initial_status

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                result = cpc.activate(
                    activation_profile_name='fake_rp',
                    force=force,
                    wait_for_completion=wait_for_completion)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            result = cpc.activate(
                activation_profile_name='fake_rp',
                force=force,
                wait_for_completion=wait_for_completion)

            if wait_for_completion:
                assert result is None
            else:
                raise NotImplementedError

            cpc.pull_full_properties()
            status = cpc.properties['status']
            assert status == exp_status

    @pytest.mark.parametrize(
        "wait_for_completion", [True]
    )
    @pytest.mark.parametrize(
        "cpc_name, initial_status, force, exp_status, exp_error", [
            # Different operational modes
            (CPC1_NAME, 'active', True, None, HTTPError_409_4),
            (CPC2_NAME, 'operating', True, 'no-power', None),
            # Different initial inactive statuses
            (CPC3_NAME, 'not-operating', False, 'no-power', None),
            (CPC3_NAME, 'not-operating', True, 'no-power', None),
            (CPC3_NAME, 'no-power', False, 'no-power', None),
            (CPC3_NAME, 'no-power', True, 'no-power', None),
            # Different initial active statuses
            (CPC3_NAME, 'operating', False, None, HTTPError_409_1),
            (CPC3_NAME, 'operating', True, 'no-power', None),
            (CPC3_NAME, 'degraded', False, None, HTTPError_409_1),
            (CPC3_NAME, 'degraded', True, 'no-power', None),
            (CPC3_NAME, 'acceptable', False, None, HTTPError_409_1),
            (CPC3_NAME, 'acceptable', True, 'no-power', None),
            (CPC3_NAME, 'exceptions', False, None, HTTPError_409_1),
            (CPC3_NAME, 'exceptions', True, 'no-power', None),
            (CPC3_NAME, 'service-required', False, None, HTTPError_409_1),
            (CPC3_NAME, 'service-required', True, 'no-power', None),
            # Different initial bad statuses
            (CPC3_NAME, 'not-communicating', False, None, HTTPError_409_1),
            (CPC3_NAME, 'not-communicating', True, None, HTTPError_409_1),
            (CPC3_NAME, 'status-check', False, None, HTTPError_409_1),
            (CPC3_NAME, 'status-check', True, None, HTTPError_409_1),
        ]
    )
    def test_cpc_deactivate(
            self, cpc_name, initial_status, force, exp_status, exp_error,
            wait_for_completion):
        """Test Cpc.deactivate()."""

        # wait_for_completion=False not implemented in mock support:
        assert wait_for_completion is True

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)

        # Set initial status of the CPC for this test
        faked_cpc.properties['status'] = initial_status

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                result = cpc.deactivate(
                    force=force,
                    wait_for_completion=wait_for_completion)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            result = cpc.deactivate(
                force=force,
                wait_for_completion=wait_for_completion)

            if wait_for_completion:
                assert result is None
            else:
                raise NotImplementedError

            cpc.pull_full_properties()
            status = cpc.properties['status']
            assert status == exp_status

    @pytest.mark.parametrize(
        "cpc_name, exp_error", [
            (CPC1_NAME, HTTPError_409_4),
            (CPC2_NAME, None),
            (CPC3_NAME, None),
        ]
    )
    def test_cpc_import_profiles(self, cpc_name, exp_error):
        """Test Cpc.import_profiles()."""

        # Add a faked CPC
        self.add_cpc(cpc_name)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)
        profile_area = 1

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                result = cpc.import_profiles(profile_area)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            result = cpc.import_profiles(profile_area)

            assert result is None

    @pytest.mark.parametrize(
        "cpc_name, exp_error", [
            (CPC1_NAME, HTTPError_409_4),
            (CPC2_NAME, None),
            (CPC3_NAME, None),
        ]
    )
    def test_cpc_export_profiles(self, cpc_name, exp_error):
        """Test Cpc.export_profiles()."""

        # Add a faked CPC
        self.add_cpc(cpc_name)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)
        profile_area = 1

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                result = cpc.export_profiles(profile_area)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            result = cpc.export_profiles(profile_area)

            assert result is None

    @pytest.mark.parametrize(
        "cpc_name, exp_error", [
            (CPC1_NAME, None),
            (CPC2_NAME, HTTPError_409_5),
            (CPC3_NAME, HTTPError_409_5),
        ]
    )
    def test_cpc_get_wwpns(self, cpc_name, exp_error):
        """Test Cpc.get_wwpns()."""

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)

        faked_fcp1 = faked_cpc.adapters.add({
            'object-id': 'fake-fcp1-oid',
            # object-uri is automatically set
            'parent': faked_cpc.uri,
            'class': 'adapter',
            'name': 'fake-fcp1-name',
            'description': 'FCP #1 in CPC #1',
            'status': 'active',
            'type': 'fcp',
            'port-count': 1,
            'adapter-id': '12F',
            # network-port-uris is automatically set when adding port
        })

        faked_port11 = faked_fcp1.ports.add({
            'element-id': 'fake-port11-oid',
            # element-uri is automatically set
            'parent': faked_fcp1.uri,
            'class': 'storage-port',
            'index': 0,
            'name': 'fake-port11-name',
            'description': 'FCP #1 Port #1',
        })

        faked_part1 = faked_cpc.partitions.add({
            'object-id': 'fake-part1-oid',
            # object-uri is automatically set
            'parent': faked_cpc.uri,
            'class': 'partition',
            'name': 'fake-part1-name',
            'description': 'Partition #1',
            'status': 'active',
        })

        faked_hba1 = faked_part1.hbas.add({
            'element-id': 'fake-hba1-oid',
            # element-uri is automatically set
            'parent': faked_part1.uri,
            'class': 'hba',
            'name': 'fake-hba1-name',
            'description': 'HBA #1 in Partition #1',
            'wwpn': 'AABBCCDDEC000082',
            'adapter-port-uri': faked_port11.uri,
            'device-number': '012F',
        })

        faked_part2 = faked_cpc.partitions.add({
            'object-id': 'fake-part2-oid',
            # object-uri is automatically set
            'parent': faked_cpc.uri,
            'class': 'partition',
            'name': 'fake-part2-name',
            'description': 'Partition #2',
            'status': 'active',
        })

        faked_hba2 = faked_part2.hbas.add({
            'element-id': 'fake-hba2-oid',
            # element-uri is automatically set
            'parent': faked_part2.uri,
            'class': 'hba',
            'name': 'fake-hba2-name',
            'description': 'HBA #2 in Partition #2',
            'wwpn': 'AABBCCDDEC000084',
            'adapter-port-uri': faked_port11.uri,
            'device-number': '012E',
        })

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)
        partitions = cpc.partitions.list()

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                wwpn_list = cpc.get_wwpns(partitions)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            wwpn_list = cpc.get_wwpns(partitions)

            exp_wwpn_list = [
                {'wwpn': faked_hba1.properties['wwpn'],
                 'partition-name': faked_part1.properties['name'],
                 'adapter-id': faked_fcp1.properties['adapter-id'],
                 'device-number': faked_hba1.properties['device-number']},
                {'wwpn': faked_hba2.properties['wwpn'],
                 'partition-name': faked_part2.properties['name'],
                 'adapter-id': faked_fcp1.properties['adapter-id'],
                 'device-number': faked_hba2.properties['device-number']},
            ]
            assert wwpn_list == exp_wwpn_list

    @pytest.mark.parametrize(
        "env_name, parm_adapter_names, exp_free_domains",
        GET_FREE_CRYPTO_DOMAINS_SUCCESS_TESTCASES)
    def test_cpc_get_free_crypto_domains(self, env_name, parm_adapter_names,
                                         exp_free_domains):
        """Test Cpc.get_free_crypto_domains()."""

        env = GET_FREE_CRYPTO_DOMAINS_ENVIRONMENTS[env_name]

        cpc_name = env['cpc_name']

        # Add the faked CPC
        faked_cpc = self.add_cpc(cpc_name)

        # Add the faked crypto adapters
        faked_adapters = {}  # faked crypto adapters by name
        for adapter_name in env['adapter_names']:
            faked_adapter = self.add_crypto_adapter(faked_cpc, adapter_name)
            faked_adapters[adapter_name] = faked_adapter

        # Add the faked partitions
        for part in env['partitions']:
            faked_part = self.add_partition(faked_cpc, part['name'])

            part_adapter_uris = [faked_adapters[name].uri
                                 for name in part['adapter_names']]
            part_domain_configs = part['domain_configs']
            crypto_config = {
                'crypto-adapter-uris': part_adapter_uris,
                'crypto-domain-configurations': part_domain_configs,
            }
            faked_part.properties['crypto-configuration'] = crypto_config

        # Set up input parameters
        cpc = self.client.cpcs.find(name=cpc_name)
        if parm_adapter_names is None:
            parm_adapters = None
        else:
            parm_adapters = [cpc.adapters.find(name=name)
                             for name in parm_adapter_names]

        # Execute the code to be tested
        act_free_domains = cpc.get_free_crypto_domains(parm_adapters)

        # Verify the result
        assert act_free_domains == exp_free_domains

    @pytest.mark.parametrize(
        "wait_for_completion", [True]
    )
    @pytest.mark.parametrize(
        "cpc_name, power_saving, exp_error", [
            (CPC1_NAME, 'high-performance', None),
            (CPC1_NAME, 'low-power', None),
            (CPC1_NAME, 'custom', None),
            (CPC1_NAME, None, HTTPError_400_7),
            (CPC1_NAME, 'invalid', HTTPError_400_7),
        ]
    )
    def test_cpc_set_power_save(self, cpc_name, power_saving, exp_error,
                                wait_for_completion):
        """Test Cpc.set_power_save()."""

        # wait_for_completion=False not implemented in mock support:
        assert wait_for_completion is True

        # Add a faked CPC
        self.add_cpc(cpc_name)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                result = cpc.set_power_save(
                    power_saving, wait_for_completion=wait_for_completion)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            result = cpc.set_power_save(
                power_saving, wait_for_completion=wait_for_completion)

            if wait_for_completion:
                assert result is None
            else:
                raise NotImplementedError

            cpc.pull_full_properties()

            assert cpc.properties['cpc-power-saving'] == power_saving
            assert cpc.properties['cpc-power-saving-state'] == power_saving
            assert cpc.properties['zcpc-power-saving'] == power_saving
            assert cpc.properties['zcpc-power-saving-state'] == power_saving

    @pytest.mark.parametrize(
        "wait_for_completion", [True]
    )
    @pytest.mark.parametrize(
        "cpc_name, power_capping_state, power_cap, exp_error", [
            (CPC1_NAME, 'disabled', None, None),
            (CPC1_NAME, 'enabled', 20000, None),
            (CPC1_NAME, 'enabled', None, HTTPError_400_7),
            (CPC1_NAME, None, None, HTTPError_400_7),
            (CPC1_NAME, 'invalid', None, HTTPError_400_7),
        ]
    )
    def test_cpc_set_power_capping(self, cpc_name, power_capping_state,
                                   power_cap, exp_error, wait_for_completion):
        """Test Cpc.set_power_capping()."""

        # wait_for_completion=False not implemented in mock support:
        assert wait_for_completion is True

        # Add a faked CPC
        self.add_cpc(cpc_name)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        if exp_error:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                result = cpc.set_power_capping(
                    power_capping_state, power_cap,
                    wait_for_completion=wait_for_completion)

            exc = exc_info.value
            assert exc.http_status == exp_error.http_status
            assert exc.reason == exp_error.reason
        else:

            # Execute the code to be tested
            result = cpc.set_power_capping(
                power_capping_state, power_cap,
                wait_for_completion=wait_for_completion)

            if wait_for_completion:
                assert result is None
            else:
                raise NotImplementedError

            cpc.pull_full_properties()

            assert cpc.properties['cpc-power-capping-state'] == \
                power_capping_state
            assert cpc.properties['cpc-power-cap-current'] == power_cap
            assert cpc.properties['zcpc-power-capping-state'] == \
                power_capping_state
            assert cpc.properties['zcpc-power-cap-current'] == power_cap

    @pytest.mark.parametrize(
        "cpc_name, energy_props", [
            (CPC1_NAME, {
                'cpc-power-consumption': 14423,
                'cpc-power-rating': 28000,
                'cpc-power-save-allowed': 'allowed',
                'cpc-power-saving': 'high-performance',
                'cpc-power-saving-state': 'high-performance',
                'zcpc-ambient-temperature': 26.7,
                'zcpc-dew-point': 8.4,
                'zcpc-exhaust-temperature': 29.0,
                'zcpc-heat-load': 49246,
                'zcpc-heat-load-forced-air': 10370,
                'zcpc-heat-load-water': 38877,
                'zcpc-humidity': 31,
                'zcpc-maximum-potential-heat-load': 57922,
                'zcpc-maximum-potential-power': 16964,
                'zcpc-power-consumption': 14423,
                'zcpc-power-rating': 28000,
                'zcpc-power-save-allowed': 'under-group-control',
                'zcpc-power-saving': 'high-performance',
                'zcpc-power-saving-state': 'high-performance',
            }),
        ]
    )
    def test_cpc_get_energy_mgmt_props(
            self, cpc_name, energy_props):
        """Test Cpc.get_energy_management_properties()."""

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)
        faked_cpc.properties.update(energy_props)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        # Execute the code to be tested
        act_energy_props = cpc.get_energy_management_properties()

        assert isinstance(act_energy_props, dict)
        cpc.pull_full_properties()

        for p in energy_props:
            exp_value = energy_props[p]

            # Verify returned energy properties
            assert p in act_energy_props
            assert act_energy_props[p] == exp_value

            # Verify consistency of returned energy properties with CPC props
            assert p in cpc.properties
            assert cpc.properties[p] == exp_value

    @pytest.mark.parametrize(
        "cpc_name, cpc_props, input_kwargs, exp_cpc_props", [
            (
                CPC1_NAME,
                CPC1_CAPACITY_PROPS,
                dict(
                    record_id='caprec1',
                ),
                CPC1_CAPACITY_PROPS,
            ),
            (
                CPC1_NAME,
                CPC1_CAPACITY_PROPS,
                dict(
                    record_id='caprec1',
                    software_model='710',
                ),
                updated(
                    CPC1_CAPACITY_PROPS, {
                        'software-model-permanent-plus-temporary': '710',
                        'processor-count-general-purpose': 10,
                    }
                ),
            ),
            (
                CPC1_NAME,
                CPC1_CAPACITY_PROPS,
                dict(
                    record_id='caprec1',
                    processor_info=dict(ifl=1),
                ),
                updated(
                    CPC1_CAPACITY_PROPS, {
                        'processor-count-ifl': 5,
                    }
                ),
            ),
        ]
    )
    def test_cpc_add_temporary_capacity(
            self, cpc_name, cpc_props, input_kwargs, exp_cpc_props):
        """Test Cpc.add_temporary_capacity()."""

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)
        faked_cpc.properties.update(cpc_props)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        # Execute the code to be tested
        cpc.add_temporary_capacity(**input_kwargs)

        cpc.pull_full_properties()

        for pname, pvalue in exp_cpc_props.items():
            assert pname in cpc.properties
            assert cpc.properties[pname] == pvalue

    @pytest.mark.parametrize(
        "cpc_name, cpc_props, input_kwargs, exp_cpc_props", [
            (
                CPC1_NAME,
                CPC1_CAPACITY_PROPS,
                dict(
                    record_id='caprec1',
                ),
                CPC1_CAPACITY_PROPS,
            ),
            (
                CPC1_NAME,
                updated(
                    CPC1_CAPACITY_PROPS, {
                        'software-model-permanent-plus-temporary': '710',
                        'processor-count-general-purpose': 10,
                    }
                ),
                dict(
                    record_id='caprec1',
                    software_model='705',
                ),
                updated(
                    CPC1_CAPACITY_PROPS, {
                        'processor-count-general-purpose': 5,
                    }
                ),
            ),
            (
                CPC1_NAME,
                CPC1_CAPACITY_PROPS,
                dict(
                    record_id='caprec1',
                    processor_info=dict(ifl=1),
                ),
                updated(
                    CPC1_CAPACITY_PROPS, {
                        'processor-count-ifl': 3,
                    }
                ),
            ),
        ]
    )
    def test_cpc_remove_temporary_capacity(
            self, cpc_name, cpc_props, input_kwargs, exp_cpc_props):
        """Test Cpc.remove_temporary_capacity()."""

        # Add a faked CPC
        faked_cpc = self.add_cpc(cpc_name)
        faked_cpc.properties.update(cpc_props)

        cpc_mgr = self.client.cpcs
        cpc = cpc_mgr.find(name=cpc_name)

        # Execute the code to be tested
        cpc.remove_temporary_capacity(**input_kwargs)

        cpc.pull_full_properties()

        for pname, pvalue in exp_cpc_props.items():
            assert pname in cpc.properties
            assert cpc.properties[pname] == pvalue

    @pytest.mark.parametrize(
        "env_name, init_asl, exp_asl", [
            (
                'env1',
                [],
                [],
            ),
            (
                'env1',
                [
                    {
                        'partition-uri': '/api/partitions/part-A-oid',
                        'post-start-delay': 10,
                    },
                ],
                [],
            ),
            (
                'env1',
                [],
                [
                    {
                        'type': 'partition',
                        'partition-uri': '/api/partitions/part-A-oid',
                        'post-start-delay': 20,
                    },
                ],
            ),
            (
                'env1',
                [
                    {
                        'type': 'partition',
                        'partition-uri': '/api/partitions/part-A-oid',
                        'post-start-delay': 10,
                    },
                ],
                [
                    {
                        'type': 'partition',
                        'partition-uri': '/api/partitions/part-A-oid',
                        'post-start-delay': 20,
                    },
                ],
            ),
            (
                'env1',
                [
                    {
                        'type': 'partition',
                        'partition-uri': '/api/partitions/part-A-oid',
                        'post-start-delay': 10,
                    },
                ],
                [
                    {
                        'type': 'partition-group',
                        'name': 'pg1',
                        'description': 'pg1 desc',
                        'partition-uris': [
                            '/api/partitions/part-A-oid',
                        ],
                        'post-start-delay': 20,
                    },
                ],
            ),
            (
                'env1',
                [
                    {
                        'type': 'partition',
                        'partition-uri': '/api/partitions/part-A-oid',
                        'post-start-delay': 10,
                    },
                ],
                [
                    {
                        'type': 'partition-group',
                        'name': 'pg1',
                        'description': 'pg1 desc',
                        'partition-uris': [
                            '/api/partitions/part-A-oid',
                        ],
                        'post-start-delay': 20,
                    },
                    {
                        'type': 'partition',
                        'partition-uri': '/api/partitions/part-B-oid',
                        'post-start-delay': 30,
                    },
                ],
            ),
        ]
    )
    def test_cpc_set_auto_start_list(
            self, env_name, init_asl, exp_asl):
        """Test Cpc.set_auto_start_list()."""
        env = SET_AUTO_START_LIST_ENVIRONMENTS[env_name]
        cpc_name = env['cpc_name']

        # Add the faked CPC
        faked_cpc = self.add_cpc(cpc_name)
        faked_cpc.properties['auto-start-list'] = init_asl

        # Add the faked partitions
        for part in env['partitions']:
            self.add_partition(faked_cpc, part['name'])

        cpc = self.client.cpcs.find(name=cpc_name)
        input_asl = []
        for asl_item in exp_asl:
            if 'partition-uri' in asl_item:
                uri = asl_item['partition-uri']
                part = cpc.partitions.find(**{'object-uri': uri})
                delay = asl_item['post-start-delay']
                input_asl.append((part, delay))
            elif 'partition-uris' in asl_item:
                uris = asl_item['partition-uris']
                parts = [p for p in cpc.partitions.list() if p.uri in uris]
                delay = asl_item['post-start-delay']
                name = asl_item['name']
                description = asl_item['description']
                input_asl.append((parts, name, description, delay))

        # Execute the code to be tested
        cpc.set_auto_start_list(input_asl)

        cpc.pull_full_properties()

        assert 'auto-start-list' in cpc.properties
        assert cpc.properties['auto-start-list'] == exp_asl

    # TODO: Test for Cpc.list_associated_storage_groups()

    # TODO: Test for Cpc.validate_lun_path()


def test_cpc_swap_cts(http_mocked_cpc_dpm):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test function for Cpc.swap_current_time_server()
    """
    uri = http_mocked_cpc_dpm.uri + '/operations/swap-cts'

    # Define the input parameters for the test call
    stp_id = 'stp-1'

    exp_request_body = {
        'stp-id': stp_id,
    }
    exp_status_code = 204

    rm_adapter = requests_mock.Adapter(case_sensitive=True)
    with requests_mock.mock(adapter=rm_adapter) as m:

        m.post(uri, status_code=exp_status_code)

        result = http_mocked_cpc_dpm.swap_current_time_server(stp_id)

        assert rm_adapter.called
        request_body = rm_adapter.last_request.json()
        assert request_body == exp_request_body
        assert result is None


def test_cpc_set_stp_config(http_mocked_cpc_dpm):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test function for Cpc.set_stp_config()
    """
    uri = http_mocked_cpc_dpm.uri + '/operations/set-stp-config'

    # Define the input parameters for the test call
    stp_id = 'stp-1'
    new_stp_id = None
    force = True
    preferred_time_server = STPNode(
        object_uri=http_mocked_cpc_dpm.uri,
        type='', model='', manuf='', po_manuf='', seq_num='',
        node_name=''
    )
    backup_time_server = None
    arbiter = None
    current_time_server = 'preferred'

    exp_request_body = {
        'stp-id': stp_id,
        'force': force,
        'preferred-time-server': preferred_time_server.json(),
        'current-time-server': current_time_server,
    }
    if new_stp_id:
        exp_request_body['new-stp-id'] = new_stp_id
    if backup_time_server:
        exp_request_body['backup-time-server'] = backup_time_server.json()
    if arbiter:
        exp_request_body['arbiter'] = arbiter.json()
    exp_status_code = 204

    rm_adapter = requests_mock.Adapter(case_sensitive=True)
    with requests_mock.mock(adapter=rm_adapter) as m:

        m.post(uri, status_code=exp_status_code)

        result = http_mocked_cpc_dpm.set_stp_config(
            stp_id, new_stp_id, force, preferred_time_server,
            backup_time_server, arbiter, current_time_server)

        assert rm_adapter.called
        request_body = rm_adapter.last_request.json()
        assert request_body == exp_request_body
        assert result is None


def test_cpc_change_stp_id(http_mocked_cpc_dpm):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test function for Cpc.change_stp_id()
    """
    uri = http_mocked_cpc_dpm.uri + '/operations/change-stponly-ctn'

    # Define the input parameters for the test call
    stp_id = 'stp-1'

    exp_request_body = {
        'stp-id': stp_id,
    }
    exp_status_code = 204

    rm_adapter = requests_mock.Adapter(case_sensitive=True)
    with requests_mock.mock(adapter=rm_adapter) as m:

        m.post(uri, status_code=exp_status_code)

        result = http_mocked_cpc_dpm.change_stp_id(stp_id)

        assert rm_adapter.called
        request_body = rm_adapter.last_request.json()
        assert request_body == exp_request_body
        assert result is None


def test_cpc_join_ctn(http_mocked_cpc_dpm):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test function for Cpc.join_ctn()
    """
    uri = http_mocked_cpc_dpm.uri + '/operations/join-stponly-ctn'

    # Define the input parameters for the test call
    stp_id = 'stp-1'

    exp_request_body = {
        'stp-id': stp_id,
    }
    exp_status_code = 204

    rm_adapter = requests_mock.Adapter(case_sensitive=True)
    with requests_mock.mock(adapter=rm_adapter) as m:

        m.post(uri, status_code=exp_status_code)

        result = http_mocked_cpc_dpm.join_ctn(stp_id)

        assert rm_adapter.called
        request_body = rm_adapter.last_request.json()
        assert request_body == exp_request_body
        assert result is None


def test_cpc_leave_ctn(http_mocked_cpc_dpm):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test function for Cpc.leave_ctn()
    """
    uri = http_mocked_cpc_dpm.uri + '/operations/leave-stponly-ctn'

    exp_status_code = 204

    rm_adapter = requests_mock.Adapter(case_sensitive=True)
    with requests_mock.mock(adapter=rm_adapter) as m:

        m.post(uri, status_code=exp_status_code)

        result = http_mocked_cpc_dpm.leave_ctn()

        assert rm_adapter.called
        assert rm_adapter.last_request.text is None
        assert result is None


TESTCASES_CPC_INSTALL_AND_ACTIVATE = [
    # Testcases for test_cpc_install_and_activate()

    # Each list item is a tuple defining a testcase in the following format:
    # - desc (str): Testcase description
    # - input_kwargs (dict): Input kwargs for method to be tested
    # - exp_request_body (dict): Expected request body for HMC operation
    # - response_status (int): HTTP status code to be returned from HMC op.
    # - response_body (dict): Response body to be returned from HMC operation
    # - exp_exc_type: Expected exception type raised from method to be tested,
    #   or None for success.

    (
        "Install all locally available updates - fail on disruptive",
        dict(
            wait_for_completion=False,
        ),
        {},
        202,
        {
            'job-uri': '/api/jobs/1',
        },
        None,
    ),
    (
        "Install all locally available updates - including disruptive",
        dict(
            install_disruptive=True,
            wait_for_completion=False,
        ),
        {
            'install-disruptive': True,
        },
        202,
        {
            'job-uri': '/api/jobs/1',
        },
        None,
    ),
    (
        "Install a bundle level - fail on disruptive",
        dict(
            bundle_level='S78',
            wait_for_completion=False,
        ),
        {
            'bundle-level': 'S78',
        },
        202,
        {
            'job-uri': '/api/jobs/1',
        },
        None,
    ),
    (
        "Install EC levels - fail on disruptive",
        dict(
            ec_levels=[('P12345', '001')],
            wait_for_completion=False,
        ),
        {
            'ec-levels': [{'number': 'P12345', 'mcl': '001'}],
        },
        202,
        {
            'job-uri': '/api/jobs/1',
        },
        None,
    ),
    (
        "Install EC levels - including disruptive",
        dict(
            ec_levels=[('P12345', '001')],
            install_disruptive=True,
            wait_for_completion=False,
        ),
        {
            'ec-levels': [{'number': 'P12345', 'mcl': '001'}],
            'install-disruptive': True,
        },
        202,
        {
            'job-uri': '/api/jobs/1',
        },
        None,
    ),
    (
        "Error: bundle-level and ec-levels specified",
        dict(
            bundle_level='S78',
            ec_levels=[('P12345', '001')],
            wait_for_completion=False,
        ),
        {},
        400,
        {   # partial error response body
            'http-status': 400,
            'reason': 1,
            'message': "bundle-level and ec-levels are both specified",
        },
        HTTPError,
    ),
]


@pytest.mark.parametrize(
    "desc, input_kwargs, exp_request_body, response_status, response_body, "
    "exp_exc_type",
    TESTCASES_CPC_INSTALL_AND_ACTIVATE)
def test_cpc_install_and_activate(
        http_mocked_cpc_dpm,  # noqa: F811
        desc, input_kwargs, exp_request_body, response_status, response_body,
        exp_exc_type):
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test function for Cpc.install_and_activate()
    """
    op_uri = http_mocked_cpc_dpm.uri + '/operations/install-and-activate'
    op_method = 'POST'

    rm_adapter = requests_mock.Adapter(case_sensitive=True)
    with requests_mock.mock(adapter=rm_adapter) as m:

        m.post(op_uri, status_code=response_status, json=response_body)

        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                http_mocked_cpc_dpm.install_and_activate(**input_kwargs)

            exc = exc_info.value
            if isinstance(exc, HTTPError):
                assert exc.http_status == response_body['http-status']
                assert exc.reason == response_body['reason']
                assert exc.message == response_body['message']
            else:
                raise AssertionError(
                    f"Unexpected exception: {exc.__class__.__name__}: {exc}")

        else:

            # The code to be tested
            result = http_mocked_cpc_dpm.install_and_activate(**input_kwargs)

            assert rm_adapter.called
            request_body = rm_adapter.last_request.json()
            assert request_body == exp_request_body
            if response_status == 202:
                # Async job started
                job_uri = response_body['job-uri']
                assert isinstance(result, Job)
                assert result.uri == job_uri
                assert result.op_method == op_method
                assert result.op_uri == op_uri
            else:
                raise AssertionError(
                    f"Unexpected HTTP status: {response_status}")


TESTCASES_CPC_DELETE_RETRIEVED_INTERNAL_CODE = [
    # Testcases for test_delete_retrieved_internal_code()

    # Each list item is a tuple defining a testcase in the following format:
    # - desc (str): Testcase description
    # - input_kwargs (dict): Input kwargs for method to be tested
    # - exp_request_body (dict): Expected request body for HMC operation
    # - response_status (int): HTTP status code to be returned from HMC op.
    # - response_body (dict): Response body to be returned from HMC operation
    # - exp_exc_type: Expected exception type raised from method to be tested,
    #   or None for success.

    (
        "Delete all retrieved but uninstalled updates",
        dict(
            wait_for_completion=False,
        ),
        {},
        202,
        {
            'job-uri': '/api/jobs/1',
        },
        None,
    ),
    (
        "Delete specific EC levels",
        dict(
            ec_levels=[('P12345', '001')],
            wait_for_completion=False,
        ),
        {
            'ec-levels': [{'number': 'P12345', 'mcl': '001'}],
        },
        202,
        {
            'job-uri': '/api/jobs/1',
        },
        None,
    ),
]


@pytest.mark.parametrize(
    "desc, input_kwargs, exp_request_body, response_status, response_body, "
    "exp_exc_type",
    TESTCASES_CPC_DELETE_RETRIEVED_INTERNAL_CODE)
def test_delete_retrieved_internal_code(
        http_mocked_cpc_dpm,  # noqa: F811
        desc, input_kwargs, exp_request_body, response_status, response_body,
        exp_exc_type):
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test function for Cpc.delete_retrieved_internal_code()
    """
    op_uri = http_mocked_cpc_dpm.uri + \
        '/operations/delete-retrieved-internal-code'
    op_method = 'POST'

    rm_adapter = requests_mock.Adapter(case_sensitive=True)
    with requests_mock.mock(adapter=rm_adapter) as m:

        m.post(op_uri, status_code=response_status, json=response_body)

        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The code to be tested
                http_mocked_cpc_dpm.delete_retrieved_internal_code(
                    **input_kwargs)

            exc = exc_info.value
            if isinstance(exc, HTTPError):
                assert exc.http_status == response_body['http-status']
                assert exc.reason == response_body['reason']
                assert exc.message == response_body['message']
            else:
                raise AssertionError(
                    f"Unexpected exception: {exc.__class__.__name__}: {exc}")

        else:

            # The code to be tested
            result = http_mocked_cpc_dpm.delete_retrieved_internal_code(
                **input_kwargs)

            assert rm_adapter.called
            request_body = rm_adapter.last_request.json()
            assert request_body == exp_request_body
            if response_status == 202:
                # Async job started
                job_uri = response_body['job-uri']
                assert isinstance(result, Job)
                assert result.uri == job_uri
                assert result.op_method == op_method
                assert result.op_uri == op_uri
            else:
                raise AssertionError(
                    f"Unexpected HTTP status: {response_status}")



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for password rules (on CPCs in DPM mode).

These tests do not change any existing password rules, but create,
modify and delete test password rules.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic PasswordRule objects (e.g. find_by_name())
PWRULE_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in PasswordRule objects returned by list() without full props
PWRULE_LIST_PROPS = ['element-uri', 'name', 'type']

# Properties whose values can change between retrievals of PasswordRule objects
PWRULE_VOLATILE_PROPS = []


def test_pwrule_find_list(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "password rules")

    # Pick the password rules to test with
    pwrule_list = console.password_rules.list()
    if not pwrule_list:
        skip_warn(f"No password rules defined on HMC {hd.host}")
    pwrule_list = pick_test_resources(pwrule_list)

    for pwrule in pwrule_list:
        print(f"Testing with password rule {pwrule.name!r}")
        runtest_find_list(
            hmc_session, console.password_rules, pwrule.name, 'name',
            'element-uri', PWRULE_VOLATILE_PROPS, PWRULE_MINIMAL_PROPS,
            PWRULE_LIST_PROPS)


def test_pwrule_property(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "password rules")

    # Pick the password rules to test with
    pwrule_list = console.password_rules.list()
    if not pwrule_list:
        skip_warn(f"No password rules defined on HMC {hd.host}")
    pwrule_list = pick_test_resources(pwrule_list)

    for pwrule in pwrule_list:
        print(f"Testing with password rule {pwrule.name!r}")

        # Select a property that is not returned by list()
        non_list_prop = 'description'

        runtest_get_properties(pwrule.manager, non_list_prop)


def test_pwrule_crud(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a password rule.
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "password rules")

    pwrule_name = TEST_PREFIX + ' test_pwrule_crud pwrule1'
    pwrule_name_new = pwrule_name + ' new'

    # Ensure a clean starting point for this test
    try:
        pwrule = console.password_rules.find(name=pwrule_name)
    except zhmcclient.NotFound:
        pass
    else:
        warnings.warn(
            f"Deleting test password rule from previous run: {pwrule_name!r}",
            UserWarning)
        pwrule.delete()

    # Test creating the password rule

    pwrule_input_props = {
        'name': pwrule_name,
        'description': 'Test password rule for zhmcclient end2end tests',
        'expiration': 90,
    }
    pwrule_auto_props = {
        'min-length': 8,
        'max-length': 256,
    }

    # The code to be tested
    try:
        pwrule = console.password_rules.create(
            pwrule_input_props)
    except zhmcclient.HTTPError as exc:
        if exc.http_status == 403 and exc.reason == 1:
            skip_warn(
                f"HMC userid {hd.userid!r} is not authorized for task "
                f"'Manage Password Rules' on HMC {hd.host}")
        else:
            raise

    for pn, exp_value in pwrule_input_props.items():
        assert pwrule.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"
    pwrule.pull_full_properties()
    for pn, exp_value in pwrule_input_props.items():
        assert pwrule.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"
    for pn, exp_value in pwrule_auto_props.items():
        assert pwrule.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"

    # Test updating a property of the password rule

    new_desc = "Updated password rule description."

    # The code to be tested
    pwrule.update_properties(dict(description=new_desc))

    assert pwrule.properties['description'] == new_desc
    pwrule.pull_full_properties()
    assert pwrule.properties['description'] == new_desc

    # Test that password rules cannot be renamed

    with pytest.raises(zhmcclient.HTTPError) as exc_info:

        # The code to be tested
        pwrule.update_properties(dict(name=pwrule_name_new))

    exc = exc_info.value
    assert exc.http_status == 400
    assert exc.reason == 6
    with pytest.raises(zhmcclient.NotFound):
        console.password_rules.find(name=pwrule_name_new)

    # Test deleting the password rule

    # The code to be tested
    pwrule.delete()

    with pytest.raises(zhmcclient.NotFound):
        console.password_rules.find(name=pwrule_name)



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for virtual switches (on CPCs in DPM mode).

These tests do not change any existing virtual switches.
"""


import pytest
from requests.packages import urllib3

# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, runtest_find_list, \
    runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic VirtualSwitch objects (e.g. find_by_name())
VSWITCH_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in VirtualSwitch objects returned by list() without full props
VSWITCH_LIST_PROPS = ['object-uri', 'name', 'type']

# Properties in VirtualSwitch objects for list(additional_properties)
VSWITCH_ADDITIONAL_PROPS = ['description', 'port']

# Properties whose values can change between retrievals of VirtualSwitch objects
VSWITCH_VOLATILE_PROPS = []


def test_vswitch_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the virtual switches to test with
        vswitch_list = cpc.virtual_switches.list()
        if not vswitch_list:
            skip_warn(
                "No virtual switches (= no OSA/HS network adapters) on "
                f"CPC {cpc.name} managed by HMC {hd.host}")
        vswitch_list = pick_test_resources(vswitch_list)

        for vswitch in vswitch_list:
            print(f"Testing on CPC {cpc.name} with virtual switch "
                  f"{vswitch.name!r}")
            runtest_find_list(
                session, cpc.virtual_switches, vswitch.name, 'name',
                'description', VSWITCH_VOLATILE_PROPS, VSWITCH_MINIMAL_PROPS,
                VSWITCH_LIST_PROPS, VSWITCH_ADDITIONAL_PROPS)


def test_vswitch_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the virtual switches to test with
        vswitch_list = cpc.virtual_switches.list()
        if not vswitch_list:
            skip_warn(
                "No virtual switches (= no OSA/HS network adapters) on "
                f"CPC {cpc.name} managed by HMC {hd.host}")
        vswitch_list = pick_test_resources(vswitch_list)

        for vswitch in vswitch_list:
            print(f"Testing on CPC {cpc.name} with virtual switch "
                  f"{vswitch.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(vswitch.manager, non_list_prop)



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Virtual Switch` is a virtualized networking switch connecting
:term:`NICs <NIC>` with a :term:`Network Port`.

Virtual Switches are generated automatically every time a new
:term:`Network Adapter` is detected and configured.

Virtual Switch resources are contained in :term:`CPC` resources.

Virtual Switches only exist in CPCs that are in DPM mode.
"""


import re
import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_VIRTUAL_SWITCH

__all__ = ['VirtualSwitchManager', 'VirtualSwitch']


class VirtualSwitchManager(BaseManager):
    """
    Manager providing access to the :term:`Virtual Switches <Virtual Switch>`
    in a particular :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Cpc` object (in DPM mode):

    * :attr:`~zhmcclient.Cpc.virtual_switches`

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, cpc):
        # This function should not go into the docs.
        # Parameters:
        #   cpc (:class:`~zhmcclient.Cpc`):
        #     CPC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',
            'type',
        ]

        super().__init__(
            resource_class=VirtualSwitch,
            class_name=RC_VIRTUAL_SWITCH,
            session=cpc.manager.session,
            parent=cpc,
            base_uri='/api/virtual-switches',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: :term:`CPC` defining the scope for this
        manager.
        """
        return self._parent

    @logged_api_call
    # pylint: disable=arguments-differ
    def list(self, full_properties=False, filter_args=None,
             additional_properties=None):
        """
        List the Virtual Switches in this CPC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to the backing Adapters of any Virtual
          Switches to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            This parameter requires HMC 2.16.0 or higher.

        Returns:

          : A list of :class:`~zhmcclient.VirtualSwitch` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'virtual-switches'
        list_uri = f'{self.cpc.uri}/virtual-switches'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args,
            additional_properties)


class VirtualSwitch(BaseResource):
    """
    Representation of a :term:`Virtual Switch`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of a Virtual Switch, see section 'Data model'
    in section 'Virtual Switch object' in the :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.VirtualSwitchManager`).

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.VirtualSwitchManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, VirtualSwitchManager), (
            "VirtualSwitch init: Expected manager type "
            f"{VirtualSwitchManager}, got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def get_connected_nics(self):
        """
        List the :term:`NICs <NIC>` connected to this Virtual Switch.

        This method performs the "Get Connected VNICs of a Virtual Switch" HMC
        operation.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to the backing Adapter of this Virtual
          Switch.

        Returns:

          : A list of :term:`Nic` objects. These objects will be connected in
          the resource tree (i.e. have a parent :term:`Partition` object,
          etc.) and will have the following properties set:

          * `element-uri`
          * `element-id`
          * `parent`
          * `class`

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.manager.session.get(
            self.uri + '/operations/get-connected-vnics', resource=self)
        nic_uris = result['connected-vnic-uris']
        nic_list = []
        parts = {}  # Key: Partition ID; Value: Partition object
        for nic_uri in nic_uris:
            m = re.match(r"^/api/partitions/([^/]+)/nics/([^/]+)/?$", nic_uri)
            part_id = m.group(1)
            nic_id = m.group(2)
            # We remember created Partition objects and reuse them.
            try:
                part = parts[part_id]
            except KeyError:
                part = self.manager.cpc.partitions.resource_object(part_id)
                parts[part_id] = part
            nic = part.nics.resource_object(nic_id)
            nic_list.append(nic)
        return nic_list

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Virtual Switch.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the backing Adapter of this Virtual
          Switch.
        * Task permission for the "Manage Adapters" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'Virtual Switch object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for storage group templates (on CPCs in DPM mode).

These tests do not change any existing storage group templates, but create,
modify and delete test storage group templates.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    skipif_no_storage_mgmt_feature, runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic StorageGroupTemplate objects (e.g. find_by_name())
STOGRPTPL_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in StorageGroupTemplate objects returned by list() without full
# props
STOGRPTPL_LIST_PROPS = ['object-uri', 'cpc-uri', 'name', 'type']

# Properties whose values can change between retrievals of StorageGroupTemplate
# objs
STOGRPTPL_VOLATILE_PROPS = []


def test_stogrptpl_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the storage group templates to test with
        stogrptpl_list = console.storage_group_templates.findall(
            **{'cpc-uri': cpc.uri})
        if not stogrptpl_list:
            skip_warn("No storage group templates associated to CPC "
                      f"{cpc.name} managed by HMC {hd.host}")
        stogrptpl_list = pick_test_resources(stogrptpl_list)

        for stogrptpl in stogrptpl_list:
            print(f"Testing on CPC {cpc.name} with storage group template "
                  f"{stogrptpl.name!r}")
            runtest_find_list(
                session, console.storage_group_templates, stogrptpl.name,
                'name', 'object-uri', STOGRPTPL_VOLATILE_PROPS,
                STOGRPTPL_MINIMAL_PROPS, STOGRPTPL_LIST_PROPS)


def test_stogrptpl_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the storage group templates to test with
        stogrptpl_list = console.storage_group_templates.findall(
            **{'cpc-uri': cpc.uri})
        if not stogrptpl_list:
            skip_warn("No storage group templates associated to CPC "
                      f"{cpc.name} managed by HMC {hd.host}")
        stogrptpl_list = pick_test_resources(stogrptpl_list)

        for stogrptpl in stogrptpl_list:
            print(f"Testing on CPC {cpc.name} with storage group template "
                  f"{stogrptpl.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(stogrptpl.manager, non_list_prop)


def test_stogrptpl_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a storage group template.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        print(f"Testing on CPC {cpc.name}")

        console = cpc.manager.client.consoles.console
        stogrptpl_name = TEST_PREFIX + ' test_stogrptpl_crud stogrptpl1'
        stogrptpl_name_new = stogrptpl_name + ' new'

        # Ensure clean starting point
        try:
            stogrptpl = console.storage_group_templates.find(
                name=stogrptpl_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test storage group template from previous run: "
                f"{stogrptpl_name!r} on CPC {cpc.name}", UserWarning)
            stogrptpl.delete()
        try:
            stogrptpl = console.storage_group_templates.find(
                name=stogrptpl_name_new)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test storage group template from previous run: "
                f"{stogrptpl_name_new!r} on CPC {cpc.name}", UserWarning)
            stogrptpl.delete()

        # Test creating the storage group template

        stogrptpl_input_props = {
            'cpc-uri': cpc.uri,
            'name': stogrptpl_name,
            'description': 'Test storage grp tpl for zhmcclient end2end tests',
            'type': 'fcp',
        }
        stogrptpl_auto_props = {
            'shared': True,
        }

        # The code to be tested
        stogrptpl = console.storage_group_templates.create(
            stogrptpl_input_props)

        for pn, exp_value in stogrptpl_input_props.items():
            assert stogrptpl.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r} of storage group " \
                f"tpl:\n{stogrptpl.properties!r}"
        stogrptpl.pull_full_properties()
        for pn, exp_value in stogrptpl_input_props.items():
            assert stogrptpl.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r} of storage group " \
                f"tpl:\n{stogrptpl.properties!r}"
        for pn, exp_value in stogrptpl_auto_props.items():
            assert pn in stogrptpl.properties, (
                f"Automatically returned property {pn!r} is not in "
                f"created storage group template:\n{stogrptpl!r}")
            assert stogrptpl.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r} of storage group " \
                f"template:\n{stogrptpl.properties!r}"

        # Test updating a property of the storage group template

        new_desc = "Updated storage group template description."

        # The code to be tested
        stogrptpl.update_properties(dict(description=new_desc))

        assert stogrptpl.properties['description'] == new_desc
        stogrptpl.pull_full_properties()
        assert stogrptpl.properties['description'] == new_desc

        # Test renaming the storage group template

        # The code to be tested
        stogrptpl.update_properties(dict(name=stogrptpl_name_new))

        assert stogrptpl.properties['name'] == stogrptpl_name_new
        stogrptpl.pull_full_properties()
        assert stogrptpl.properties['name'] == stogrptpl_name_new
        with pytest.raises(zhmcclient.NotFound):
            console.storage_group_templates.find(name=stogrptpl_name)

        # Test deleting the storage group template

        # The code to be tested
        stogrptpl.delete()

        with pytest.raises(zhmcclient.NotFound):
            console.storage_group_templates.find(name=stogrptpl_name_new)
